/**
* The contents of this file are subject to the Mozilla Public 
* License Version 1.1 (the "License"); you may not use this 
* file except in compliance with the License. You may obtain a
* copy of the License at http://www.mozilla.org/MPL/
*
* Software distributed under the License is distributed on an 
* "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express 
* or implied. See the License for the specific language 
* governing rights and limitations under the License.
*
* The Original Code is this file as it was released on
* March 8, 2006.
*
* The Initial Developer of the Original Code is Terry Yuen. 
* Portions created by Terry Yuen are Copyright (C) 2006.
* All Rights Reserved.
**/

function DictServer() {
  this.init();
}

DictServer.prototype = {
  SP_CORRECT           : 1,  //output responses
  SP_INCORRECT_AVAIL   : 2,
  SP_INCORRECT_NOAVAIL : 3,
  SP_SERVER_ERROR      : 4,
  SP_CONNECTION_ERROR  : 5,

  C_UNREADY            : 1,  //220 ..send CLIENT
  C_READY              : 2,
  C_SENT_MATCH_EXACT   : 3,
  C_SENT_MATCH_NEAR    : 4,
  C_RECEIVING_NEAR     : 5,
  C_RECEIVED_NEAR      : 6,
  C_ENDED              : 7,

  callback: null,
  response: {
    code: 0,
    text: "",  
    suggestions: null,
  },

  host: "",  //e.g., "dict.org"
  port: 2628,
  database: "! ",
  match: ". ",

  socket: null,
  input: null,
  output: null,

  state: 1,

  init: function() {
    this._inStreamPump.r = this;
  },

  query: function(str, callback) {
    if(this.response.text == str && 
       this.response.code != this.SP_SERVER_ERROR &&
       this.response.code != this.SP_CONNECTION_ERROR) {
      callback(this.response);
      return;
    }
    this.response.code = this.SP_CONNECTION_ERROR;  //safe default
    this.response.text = str;
    this.response.suggestions = null;
    this.callback = callback;

    this.state = this.C_UNREADY;
    this.connect();
    //this.response.suggestions = ["waterhouse", "coopers"];
    //this.response.code = this.SP_INCORRECT_AVAIL;
    //setTimeout(function(d){d.callback(d.response);},1000,this);
  },

  returnResult: function() {
    this.callback(this.response);
    this.callback = null;
  },

  handleError: function(c, s) {
    if(parseInt(s.substring(0, 3)) >= 400) {
      this.disconnect();
      this.response.code = c;
      this.response.text = s;
      this.returnResult();
    }
  },

  connect: function() {
    try {
      this.disconnect();  //a precaution

      var socket = Components.classes["@mozilla.org/network/socket-transport-service;1"]
          .getService(Components.interfaces.nsISocketTransportService)
          .createTransport(null, 0, this.host, this.port, null);

      this.output = socket.openOutputStream(0, 0, 0);

      var _input = socket.openInputStream(0, 0, 0);
      this.input = Components.classes["@mozilla.org/scriptableinputstream;1"]
          .createInstance(Components.interfaces.nsIScriptableInputStream);
      this.input.init(_input);

      var pump = Components.classes["@mozilla.org/network/input-stream-pump;1"].
                 createInstance(Components.interfaces.nsIInputStreamPump);
      pump.init(_input, -1, -1, 0, 0, false);
      pump.asyncRead(this._inStreamPump, null);

    } catch(e) {
      this.disconnect();
      this.handleError(this.SP_CONNECTION_ERROR, "520 No Connection");
    }
  },

  _inStreamPump: {
    buffer: "",
    r: null,
    onDataAvailable: function(request, context, inputStream, offset, count) {
      try {
        this.buffer += this.r.input.read(count);
      } catch(e) {
        this.r.handleError(this.r.SP_SERVER_ERROR, "520 IO Error");
        dump("\nRiteOfTongue stroke:" + e);
        return;
      }

      var i;
      while((i = this.buffer.indexOf("\n")) >= 0) {
        var tmp = this.buffer.substring(0, i);
        this.buffer = this.buffer.substring(i + 1, this.buffer.length);
        this.r.process(tmp);
      }
    },
    onStartRequest: function(request, context) {},
    onStopRequest: function(request, context, status) {
      this.buffer = "";
    }
  },

  process: function(str) {
    var code = str.substring(0, 3);  //may be malformed

    switch(this.state) {

      case this.C_UNREADY :
        if(code == "220") {  //initial connection
          //pipeline the output
          var str = "CLIENT RiteOfTongue\r\n";
          str += "MATCH " + this.database + "exact '" + this.response.text + "'\r\n";
          str += "MATCH " + this.database + this.match + "'" + this.response.text + "'\r\n";
          str += "QUIT\r\n";

          this.output.write(str, str.length);
          this.state = this.C_READY;

        } else {
          this.handleError(this.SP_SERVER_ERROR, str);
        }
        break;

      case this.C_READY :
        if(code == "250") {  //ok
          this.state = this.C_SENT_MATCH_EXACT;

        } else {
          this.handleError(this.SP_SERVER_ERROR, str);
        }
        break;

      case this.C_SENT_MATCH_EXACT :
        if(code == "552") {  //no match
          this.state = this.C_SENT_MATCH_NEAR;

        } else if(code == "152") {  //1 match = correct word
          this.state = this.C_ENDED;
          this.response.code = this.SP_CORRECT;
          this.returnResult();

        } else {
          this.handleError(this.SP_SERVER_ERROR, str);
        }
        break;

      case this.C_SENT_MATCH_NEAR :
        if(code == "152") {  //n matches = suggestions
          this.state = this.C_RECEIVING_NEAR;
        } else if(code == "552") {  //no match = no suggestions
          this.state = this.C_ENDED;
          this.response.code = this.SP_INCORRECT_NOAVAIL;
          this.returnResult();

        } else {
          this.handleError(this.SP_SERVER_ERROR, str);
        }
        break;

      case this.C_RECEIVING_NEAR :
        if(str.charAt(0) == "." && str.charAt(1) != ".") {  //done
          this.state = this.C_RECEIVED_NEAR;
        } else {
          str = str.substring(str.indexOf(" ") + 1, str.length);

          if(str.charAt(str.length - 1) == "\r")
            str = str.substring(0, str.length - 1);

          if(str.charAt(0) == "\"" || str.charAt(0) == "'")
            str = str.substring(1, str.length - 1);

          if(this.response.suggestions == null)this.response.suggestions = [];
          this.response.suggestions.push(str);
        }
        break;

      case this.C_RECEIVED_NEAR :
        if(code == "250") {  //ok
          this.state = this.C_ENDED;
          this.response.code = this.SP_INCORRECT_AVAIL;
          this.returnResult();
        } else {
          this.handleError(this.SP_SERVER_ERROR, str);
        }
        break;

      case this.C_ENDED :
        if(code == "221") {  //closing connection
          this.disconnect();
        }
        break;
    }
  },

  disconnect: function() {
    try {
      if(this.input)this.input.close();
    } finally {
      this.input = null;
    }

    try {
      if(this.output)this.output.close();
    } finally {
      this.output = null;
    }
  }
};