/**
* The contents of this file are subject to the Mozilla Public 
* License Version 1.1 (the "License"); you may not use this 
* file except in compliance with the License. You may obtain a
* copy of the License at http://www.mozilla.org/MPL/
*
* Software distributed under the License is distributed on an 
* "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express 
* or implied. See the License for the specific language 
* governing rights and limitations under the License.
*
* The Original Code is this file as it was released on
* March 17, 2005.
*
* The Initial Developer of the Original Code is Terry Yuen. 
* Portions created by Terry Yuen are Copyright (C) 2005.
* All Rights Reserved.
**/

function TongueSpeaker() {
}

TongueSpeaker.prototype = {
  STATE_WAIT: 1,
  STATE_CORRECT: 2,
  STATE_INCORRECT: 3,
  STATE_LIST: 4,
  STATE_ERROR: 5,
  STATE_DORMANT: 6,

  pref: null,
  showInSubmenu: true,
  selectOnly: false,

  server: new DictServer(),
  fixCapitals: false,
  displayLimit: 6,

  serverList: null,
  serverRotate: 1,

  contextMenu: null,
  input: null,
  data: null,

  localeStrings: {
    "submenu.suggest": null, 
    "accesskey.suggest": null, 
    "menu.wait": null, 
    "menu.correct": null, 
    "menu.wrong": null
  },

  startup: function() {
    var bundle = Components.classes["@mozilla.org/intl/stringbundle;1"]
        .getService(Components.interfaces.nsIStringBundleService)
        .createBundle("chrome://riteoftongue/locale/browser.properties");

    for(var s in this.localeStrings) {
      this.localeStrings[s] = bundle.GetStringFromName(s);
    }

    this.pref = Components.classes["@mozilla.org/preferences-service;1"]
        .getService(Components.interfaces.nsIPrefService)
        .getBranch("riteoftongue.");

    this.refreshPreferences();
  },

  refreshPreferences: function(uninit) {
    if(uninit) {
      this.unhookUpMenu("urlbar");
      this.unhookUpMenu("searchbar");
      this.unhookUpMenu("contentAreaContextMenu");
    }
    try {
      this.showInSubmenu = this.pref.getBoolPref("use_submenu");
    } catch(e) {
      this.showInSubmenu = true;
    }
    this.selectOnly = this.pref.getBoolPref("selections_only");
    this.fixCapitals = this.pref.getBoolPref("fix_capitalization");
    this.displayLimit = this.pref.getIntPref("display_limit");

    this.serverList = this.pref.getCharPref("dict_servers").split(" ");
    this.serverRotate = this.pref.getIntPref("dict_servers_rotate") % this.serverList.length;

    this.hookUpMenu("contentAreaContextMenu");
    if(this.pref.getBoolPref("show_in_urlbar")) {
      this.hookUpMenu("urlbar");
      this.hookUpMenu("searchbar");
    }
  },

  hookUpMenu: function(id) {
    var elm = document.getElementById(id);
    if(elm) {
      elm.addEventListener("popupshowing", RiteOfTongueOnPopupShowing, false);
      elm.addEventListener("popuphiding", RiteOfTongueOnPopupHiding, false);
    }
  },

  unhookUpMenu: function(id) {
    var elm = document.getElementById(id);
    if(elm) {
      elm.removeEventListener("popupshowing", RiteOfTongueOnPopupShowing, false);
      elm.removeEventListener("popuphiding", RiteOfTongueOnPopupHiding, false);
    }
  },

  onPopupShowing: function(event) {
    var target = event.originalTarget;

    if(!(target instanceof XULElement) || 
        (target.localName != "menupopup" && 
         target.localName != "popup")) {
      target = event.explicitOriginalTarget;
    }

    if(!this.contextMenu) {
      this.contextMenu = target;

    } else if(!target.hasAttribute("tongue")) {
      return true;  //Cancel if the event bubbled out from a child
    }

    if(!this.showInSubmenu || target.hasAttribute("tongue")) {  //show immediate

      if(!target.hasAttribute("tongue")) {
        //we need this data to determine whether to show the submenu
        this.input = this.getFormField();
        this.data = (this.input) ? this.extractInput(this.input) : null;
      }

      this.ensureMenus();
      this.clearOldEntries(target);

      if(this.data) {
        this.updateStatusMenus(this.STATE_WAIT);

        const dis = this;
        this.prepareServer();
        this.server.query(this.data.value, function(resp) {
          if(resp.code == dis.server.SP_CORRECT) {
            dis.updateStatusMenus(dis.STATE_CORRECT);

          } else if(resp.code == dis.server.SP_INCORRECT_NOAVAIL) {
            dis.updateStatusMenus(dis.STATE_INCORRECT);

          } else if(resp.code == dis.server.SP_INCORRECT_AVAIL) {
            dis.updateStatusMenus(dis.STATE_LIST);
            dis.clearOldEntries(target);
            dis.displayEntries(resp.suggestions);

          } else if(resp.code == dis.server.SP_SERVER_ERROR || 
                    resp.code == dis.server.SP_CONNECTION_ERROR) {
            dis.updateStatusMenus(dis.STATE_ERROR, resp.text);
          }
        });
      } else {
        this.updateStatusMenus(this.STATE_DORMANT);
      }

    } else {
      //if the root menu is shown, show/hide submenus
      this.input = this.getFormField();
      this.data = (this.input) ? this.extractInput(this.input) : null;

      this.ensureMenus();
      this.clearOldEntries(target);
      this.updateStatusMenus((this.data) ? this.STATE_WAIT : this.STATE_DORMANT);
    }
  },

  onPopupHiding: function(event) {
    var target = event.originalTarget;

    if(!(target instanceof XULElement) || 
        (target.localName != "menupopup" && 
         target.localName != "popup")) {
      target = event.explicitOriginalTarget;
    }

    if(target == this.contextMenu) {
      this.contextMenu = null;
      this.input = null;
      this.data = null;
    }
  },

  getFormField: function() {
    var input = document.popupNode;

    if(input instanceof HTMLInputElement || 
       input instanceof HTMLTextAreaElement) {

      input = new XPCNativeWrapper(
          input, "selectionStart", "selectionEnd", "value", 
          "setSelectionRange()", "getAttribute()", "QueryInterface()"
      );

      if(input.getAttribute("type") == "password")return null;

      return input;

    } else {

      input = new XPCNativeWrapper(
          input, "ownerDocument", "execCommand()"
      );

      if(input.ownerDocument.designMode != "on")return null;

      input.iframe = true;

      return input;

    } return null;
  },

  extractInput: function(input) {
    if(input.iframe) {
      return this.extractInputFromIframe(input);
    } else {
      return this.extractInputFromTextarea(input);
    }
  },

  extractInputFromTextarea: function(input) {
    var data = {
      iframe: false,
      start: input.selectionStart,
      end: input.selectionEnd,
      whiteBefore: "",
      whiteAfter: ""
    };

    //if the user has already selected the word for us
    if(data.start < data.end) {
      if(data.end - data.start > 25)return null;  //quick performance check

      data.value = input.value.substring(data.start, data.end);

      //save whitespaces surrounding it and check alphanumeric
      var exp = data.value.match(/^(\s*)([\w\-]+)(\s*)$/);
      if(!exp)return null;  //abort if contain spaces in word: "morgan stanley"

      data.whiteBefore = exp[1];
      data.value = exp[2];
      data.whiteAfter = exp[3];

      return data;

    } else if(!this.selectOnly && data.start != -1) {  //gather text under caret

      const isChar = /[\w\-]/;
      var val = input.value;
      var len = val.length;
      while(data.start > 0 && isChar.test(val.charAt(data.start - 1)))data.start--;
      while(data.end < len && isChar.test(val.charAt(data.end)))data.end++;

      //make sure at least 1 letter exists
      if(data.start == data.end)return null;

      data.value = val.substring(data.start, data.end);
      return data;

    } else return null;
  },

  extractInputFromIframe: function(input) {
    var data = {
      whiteBefore: "",
      whiteAfter: ""
    };

    var sel = new XPCNativeWrapper(input.ownerDocument, "defaultView").defaultView;
    sel = new XPCNativeWrapper(sel, "getSelection()").getSelection();
    data.value = sel.toString();
    sel = sel.getRangeAt(0);

    if(!sel.collapsed) {
      //save whitespaces surrounding it and check alphanumeric
      var exp = data.value.match(/^(\s*)([\w\-]+)(\s*)$/);
      if(!exp)return null;  //abort if contain spaces inside word: "morgan stanley"

      data.whiteBefore = exp[1];
      data.value = exp[2];
      data.whiteAfter = exp[3];

      return data;

    } else if(!this.selectOnly && sel.startContainer.nodeType == Node.TEXT_NODE) {

      const isChar = /[\w\-]/;
      data.start = sel.startOffset;
      data.end = sel.endOffset;
      var val = sel.startContainer.nodeValue;
      var len = val.length;
      while(data.start > 0 && isChar.test(val.charAt(data.start - 1)))data.start--;
      while(data.end < len && isChar.test(val.charAt(data.end)))data.end++;

      //make sure at least 1 letter exists
      if(data.start == data.end)return null;

      data.value = val.substring(data.start, data.end);
      return data;

    } else return null;
  },

  onCommand: function(event) {
    var target = event.originalTarget;

    if(target.getAttribute("tongue") == "suggestion") {
      if(this.input.iframe) {

        var sel = new XPCNativeWrapper(this.input.ownerDocument, "defaultView").defaultView;
        sel = new XPCNativeWrapper(sel, "getSelection()").getSelection();
        sel = sel.getRangeAt(0);
        if(sel.collapsed) {
          //select the text in the iframe
          sel.setStart(sel.startContainer, this.data.start);
          sel.setEnd(sel.endContainer, this.data.end);
        }
        var newStr = target.getAttribute("label");
        this.input.ownerDocument.execCommand("inserthtml", false, this.data.whiteBefore + newStr + this.data.whiteAfter);

      } else {
        //save scroll position
        var doc=null, docTop=0, docLeft=0;
        if("nsIDOMNSEditableElement" in Components.interfaces) {  //Firefox 1.5+
          doc = this.input.QueryInterface(Components.interfaces.nsIDOMNSEditableElement).editor.rootElement;
          docTop = doc.scrollTop;
          docLeft = doc.scrollLeft;
        }

        //recover the text
        var oldStr = this.input.value;
        var newStr = target.getAttribute("label");
        this.input.value = oldStr.substring(0, this.data.start) + 
            this.data.whiteBefore + 
            newStr + 
            this.data.whiteAfter + 
            oldStr.substring(this.data.end, oldStr.length);
        var tmp = this.data.start + newStr.length;
        this.input.setSelectionRange(tmp, tmp);  //position caret at end

        //restore scroll position
        if(doc) {  //Firefox 1.5+
          doc.scrollTop = docTop;
          doc.scrollLeft = docLeft;
        }
      }
    }
  },

  ensureMenus: function() {
    //menus known to already exist
    if(this.contextMenu.id == "contentAreaContextMenu") {
      return;
    }

    //loop to see if menus already exist
    var elm = this.contextMenu.lastChild;
    while(elm && elm.hasAttribute("tongue")) {
      if(elm.localName == "menu") {
        return;
      }
      elm = elm.previousSibling;
    }

    //otherwise create them
    var frag = document.createDocumentFragment();

    elm = document.createElement("menuseparator");
    elm.setAttribute("tongue", "system");
    frag.appendChild(elm);

    elm = document.createElement("menu");
    elm.setAttribute("tongue", "system");
    elm.setAttribute("label", this.localeStrings["submenu.suggest"]);
    elm.setAttribute("accesskey", this.localeStrings["submenu.accesskey"]);
    frag.appendChild(elm);

    elm = document.createElement("menupopup");
    elm.setAttribute("tongue", "system");
    frag.lastChild.appendChild(elm);

    elm = document.createElement("menuitem");
    elm.setAttribute("tongue", "wait");
    elm.setAttribute("label", this.localeStrings["menu.wait"]);
    elm.setAttribute("class", "menuitem-iconic");
    elm.setAttribute("disabled", "true");
    frag.appendChild(elm);

    this.contextMenu.appendChild(frag);
  },

  clearOldEntries: function(target) {
    var elm = target.lastChild;
    var t = null;
    while(elm && elm.hasAttribute("tongue")) {
      t = elm;
      elm = elm.previousSibling;
      if(t.getAttribute("tongue") == "suggestion" || t.hasAttribute("transient")) {
        target.removeChild(t);
      }
    }
  },

  displayEntries: function(suggestions) {
    if(this.fixCapitals) {
      var title = this.data.value.charAt(0);
      title = (title == title.toUpperCase());
      for(var i=0; i<suggestions.length && i<this.displayLimit; i++) {
        if(title) {
          suggestions[i] = suggestions[i].charAt(0).toUpperCase() + suggestions[i].substring(1, suggestions[i].length);
        } else {
          suggestions[i] = suggestions[i].toLowerCase();
        }
      }
    }

    var menu = this.contextMenu;

    //we want to get the actual popup menu to populate
    if(this.showInSubmenu) {
      if(menu.id == "contentAreaContextMenu") {
        menu = document.getElementById("context-riteoftongue-popup");
      } else {
        menu = menu.lastChild;
        while(menu && menu.hasAttribute("tongue")) {
          if(menu.localName == "menu") {
            menu = menu.firstChild;
            break;
          }
          menu = menu.previousSibling;
        }
      }
    }

    var frag = document.createDocumentFragment();
    var elm = null;
    //if(menu.hasChildNodes()) {
    //  elm = document.createElement("menuseparator");
    //  elm.setAttribute("tongue", "system");
    //  elm.setAttribute("transient", "true");
    //  frag.appendChild(elm);
    //}
    for(var i=0; i<suggestions.length && i<this.displayLimit; i++) {
      elm = document.createElement("menuitem");
      elm.setAttribute("label", suggestions[i]);
      elm.setAttribute("tongue", "suggestion");
      elm.setAttribute("oncommand", "RiteOfTongueOnCommand(event);");
      frag.appendChild(elm);
    }
    menu.appendChild(frag);
  },

  updateStatusMenus: function(state, errorStr) {
    if(this.contextMenu.id == "contentAreaContextMenu") {

    var _get = function(s) {
      return document.getElementById(s);
    };

      var menu = _get("context-riteoftongue-menu");
      if(this.showInSubmenu && (state == this.STATE_WAIT || state == this.STATE_LIST)) {
        menu.collapsed = false;
        menu.firstChild.collapsed = false;
      } else {
        if(menu.firstChild.hidePopup)menu.firstChild.hidePopup();
        menu.firstChild.collapsed = true;
        menu.collapsed = true;
      }
      _get("context-riteoftongue-separator").hidden = (state == this.STATE_DORMANT);

      //set the corresponding status message
      var elm = _get("context-riteoftongue-status");
      elm.hidden = (state == this.STATE_DORMANT || state == this.STATE_LIST || (state == this.STATE_WAIT && this.showInSubmenu));
      switch(state) {
        case this.STATE_WAIT:
          if(!this.showInSubmenu) {
            elm.setAttribute("tongue", "wait");
            elm.setAttribute("label", this.localeStrings["menu.wait"]);
          }
          break;

        case this.STATE_CORRECT:
          elm.setAttribute("tongue", "correct");
          elm.setAttribute("label", this.localeStrings["menu.correct"]);
          break;

        case this.STATE_INCORRECT:
          elm.setAttribute("tongue", "wrong");
          elm.setAttribute("label", this.localeStrings["menu.wrong"]);
          break;

        case this.STATE_ERROR:
          elm.setAttribute("tongue", "error");
          elm.setAttribute("label", errorStr);
          break;
      }

      //must create wait menu dynamically
      if(this.showInSubmenu && state == this.STATE_WAIT) {
        this.clearOldEntries(menu.firstChild);
        elm = document.createElement("menuitem");
        elm.setAttribute("tongue", "wait");
        elm.setAttribute("label", this.localeStrings["menu.wait"]);
        elm.setAttribute("disabled", "true");
        elm.setAttribute("transient", "true");
        menu.firstChild.appendChild(elm);
      }

      return;
    }

    //else loop through whole menu
    elm = this.contextMenu.lastChild;
    while(elm && elm.hasAttribute("tongue")) {
      switch(elm.getAttribute("tongue")) {
        case "wait":
        case "correct":
        case "wrong":
        case "error":
          elm.hidden = (state == this.STATE_DORMANT || state == this.STATE_LIST || (state == this.STATE_WAIT && this.showInSubmenu));
          switch(state) {
            case this.STATE_WAIT:
              if(!this.showInSubmenu) {
                elm.setAttribute("tongue", "wait");
                elm.setAttribute("label", this.localeStrings["menu.wait"]);
              }
              break;

            case this.STATE_CORRECT:
              elm.setAttribute("tongue", "correct");
              elm.setAttribute("label", this.localeStrings["menu.correct"]);
              break;

            case this.STATE_INCORRECT:
              elm.setAttribute("tongue", "wrong");
              elm.setAttribute("label", this.localeStrings["menu.wrong"]);
              break;

            case this.STATE_ERROR:
              elm.setAttribute("tongue", "error");
              elm.setAttribute("label", this.localeStrings["menu.error"]);
              break;

          }
          break;

        case "system":
          if(elm.localName == "menu") {
            if(this.showInSubmenu && (state == this.STATE_WAIT || state == this.STATE_LIST)) {
              elm.collapsed = false;
              elm.firstChild.collapsed = false;
            } else {
              if(elm.firstChild.hidePopup)elm.firstChild.hidePopup();
              elm.firstChild.collapsed = true;
              elm.collapsed = true;
            }
            //while we're at it, create the wait menu if needed
            if(this.showInSubmenu && state == this.STATE_WAIT) {
              this.clearOldEntries(elm.firstChild);
              var elm2 = document.createElement("menuitem");
              elm2.setAttribute("tongue", "wait");
              elm2.setAttribute("label", this.localeStrings["menu.wait"]);
              elm2.setAttribute("disabled", "true");
              elm2.setAttribute("transient", "true");
              elm.firstChild.appendChild(elm2);  //nebyoioyo
            }
          } else {  //menuseparator
            elm.hidden = (state == this.STATE_DORMANT);
          }
          break;
      }

      elm = elm.previousSibling;
    }
  },

  prepareServer: function() {
    this.serverRotate++;
    this.serverRotate %= this.serverList.length;

    var server = this.serverList[this.serverRotate];
    if(server.indexOf(":") == -1) {
      this.server.host = server;
      this.server.port = 2628;
    } else {
      server = server.split(":");
      this.server.host = server[0];
      this.server.port = server[1];
    }

    this.pref.setIntPref("dict_servers_rotate", this.serverRotate);
  },

  destroy: function() {
    this.pref = null;
    this.server = null;
    this.localeStrings = null;
    this.contextMenu = null;
    this.input = null;
    this.unhookUpMenu("urlbar");
    this.unhookUpMenu("searchbar");
    this.unhookUpMenu("contentAreaContextMenu");
  }
};

var RiteOfTongue = new TongueSpeaker();

function RiteOfTongueStartup() {RiteOfTongue.startup();}
function RiteOfTongueDestroy() {RiteOfTongue.destroy();}
function RiteOfTongueOnPopupShowing(event) {RiteOfTongue.onPopupShowing(event);}
function RiteOfTongueOnPopupHiding(event) {RiteOfTongue.onPopupHiding(event);}
function RiteOfTongueOnCommand(event) {RiteOfTongue.onCommand(event);}

window.addEventListener("load", RiteOfTongueStartup, false);
window.addEventListener("unload", RiteOfTongueDestroy, false);